import logging

logger = logging.getLogger("simput.core.domains")
logger.setLevel(logging.INFO)

# -----------------------------------------------------------------------------
# PropertyDomain
# -----------------------------------------------------------------------------


class PropertyDomain:
    """
    A Domain is responsible for:
      - testing if a property is valid
      - compute and set a property value
      - list what are the possible value options

     Domain meta information:
     - level describe the importance of the domain
        (0: info), (1: warning), (2: error)
     - message personalize the context in which a given domain is used to help
       understand why a given value is not valid.
    """

    def __init__(self, _proxy, _property: str, **kwargs):
        self._proxy = _proxy
        self._property_name = _property
        self._dependent_properties = set([self._property_name])
        self._need_set = "initial" in kwargs
        self._level = kwargs.get("level", 0)
        self._message = kwargs.get("message", str(__class__))
        self._should_compute_value = "initial" in kwargs

    def __del__(self):
        logger.info(
            "PropertyDomain::__del__ %s[%s].%s",
            self._proxy.type,
            self._proxy.id,
            self._property_name,
        )

    def enable_set_value(self):
        """Reset domain set so it can re-compute a default value"""
        self._should_compute_value = True

    def set_value(self):
        """
        Ask domain to compute and set a value to a property.
        return True if the action was successful.
        """
        return False

    def available(self):
        """List the available options"""
        return None

    @property
    def value(self):
        """Return the current proxy property value on which the domain is bound"""
        return self._proxy[self._property_name]

    @value.setter
    def value(self, v):
        """Set the proxy property value"""
        self._proxy.set_property(self._property_name, v)

    def valid(self, required_level=2):
        """Return true if the current proxy property value is valid for the given level"""
        return True

    @property
    def level(self):
        """Return current domain level (0:info, 1:warn, 2:error)"""
        return self._level

    @level.setter
    def level(self, value):
        """Update domain level"""
        self._level = value

    @property
    def message(self):
        """Associated domain message that is used for hints"""
        return self._message

    @message.setter
    def message(self, value):
        """Update domain message"""
        self._message = value

    def hints(self):
        """Return a set of (level, message) when running the validation for the info level"""
        if self.valid(-1):
            return []
        return [
            {
                "level": self._level,
                "message": self._message,
            }
        ]


# -----------------------------------------------------------------------------
# LabelList
# -----------------------------------------------------------------------------
#  name: xxxx      : (optional) provide another name than its type
# -----------------------------------------------------------------------------
#  values: [{ text, value}, ...]
# -----------------------------------------------------------------------------


class LabelList(PropertyDomain):
    def __init__(self, _proxy, _property: str, **kwargs):
        super().__init__(_proxy, _property, **kwargs)
        self._values = kwargs.get("values", [])
        self._message = "LabelList"

    def set_value(self):
        if self._should_compute_value and self._values:
            self._should_compute_value = False
            # first
            self.value = self._values[0].get("value")
            return True
        return False

    def available(self):
        return self._values

    def valid(self, required_level=2):
        if self._level < required_level:
            return True

        v = self.value
        for item in self._values:
            if item.get("value", None) == v:
                return True
        return False


# -----------------------------------------------------------------------------
# Range
# -----------------------------------------------------------------------------
#  name: xxxx                 (optional) provide another name than its type
# -----------------------------------------------------------------------------
#  value_range: [0, 1]        Static range
# -----------------------------------------------------------------------------
#  property: PropArray        Specify property on which an array is defined
#  initial: [mean, min, max]  Computation to use for setting the value
#  component: -1 (mag)        Component to use for range computation
# -----------------------------------------------------------------------------


class Range(PropertyDomain):
    def __init__(self, _proxy, _property: str, **kwargs):
        super().__init__(_proxy, _property, **kwargs)
        self.__compute = kwargs.get("initial", "mean")
        self.__prop_array = kwargs.get("property", None)
        self.__range_component = kwargs.get("component", -1)
        self.__static_range = kwargs.get("value_range", False)
        self._message = "Range"

        # Declare dependency
        if self.__prop_array:
            # ic("Range add dep", self.__prop_array)
            self._dependent_properties.add(self.__prop_array)

    def set_value(self):
        if self._should_compute_value:
            data_range = self.get_range(self.__range_component)
            if data_range is None:
                return False

            _v = 0
            if self.__compute == "mean":
                _v = (data_range[0] + data_range[1]) * 0.5
            elif self.__compute == "min":
                _v = data_range[0]
            elif self.__compute == "max":
                _v = data_range[1]
            else:
                print(
                    f"Range domain can't compute {self.__compute}. Expect 'mean', 'min' or 'max' instead."
                )

            prop_size = self._proxy.definition.get(self._property_name).get("size", 1)
            if prop_size == -1:
                self.value = [_v]
            elif prop_size > 1:
                if self.value and isinstance(self.value, list):
                    if len(self.value):
                        self.value[0] = _v
                    else:
                        self.value.append(_v)
                else:
                    self.value = [_v]
            else:
                self.value = _v

            self._should_compute_value = False
            return True
        return False

    def available(self):
        return self.get_range(self.__range_component)

    def valid(self, required_level=2):
        if self._level < required_level:
            return True

        _v = self.value
        if _v is None:
            self._message = (
                f"Undefined value can not be evaluated in {self.available()}"
            )
            return False

        _range = self.available()
        self._message = f"Value outside of {_range}"
        if isinstance(_v, list):
            _valid = True
            for __v in _v:
                if __v is None:
                    continue
                if _range[0] is not None and __v < _range[0]:
                    _valid = False
                if _range[1] is not None and __v > _range[1]:
                    _valid = False
            return _valid

        lower = _range[0] is None or _v >= _range[0]
        upper = _range[1] is None or _v <= _range[1]
        return lower and upper

    def get_range(self, component=-1):
        if self.__static_range:
            return self.__static_range

        return None


# -----------------------------------------------------------------------------
# Factory management of PropertyDomain
# -----------------------------------------------------------------------------


REGISTERED_DOMAIN_BY_TYPE = {}


def create_property_domain(proxy, proxy_prop_name, type=None, **domain_config):
    if type in REGISTERED_DOMAIN_BY_TYPE:
        domain = REGISTERED_DOMAIN_BY_TYPE[type]
        if domain:
            return domain(proxy, proxy_prop_name, **domain_config)
        return False


def register_property_domain(domain_type, domain_class):
    global REGISTERED_DOMAIN_BY_TYPE
    REGISTERED_DOMAIN_BY_TYPE[domain_type] = domain_class


# -----------------------------------------------------------------------------
# Default Skip
# -----------------------------------------------------------------------------


SKIP_NAMES = [
    "PropertyList",
    "Boolean",
    "UI",
]
for name in SKIP_NAMES:
    register_property_domain(name, False)


# -----------------------------------------------------------------------------
# Default initialization
# -----------------------------------------------------------------------------


register_property_domain("LabelList", LabelList)
register_property_domain("Range", Range)



def create_id_generator(prefix=""):
    """Generic id generator"""
    count = 1
    while True:
        yield f"{prefix}{count}"
        count += 1


def is_equal(a, b):
    """Return True if both value can be conciderated as the same"""
    # might need some deeper investigation
    return a == b


def is_valid_value(v):
    """Return true if a value can be stored into a property of a proxy"""
    if v is None:
        return False
    if isinstance(v, (str, bool, int, float)):
        return True
    if isinstance(v, (list, tuple)):
        for item in v:
            if not is_valid_value(item):
                return False
        return True
    return False



from pathlib import Path
from .protocol import SimputProtocol

# Compute local path to serve
serve_path = str(Path(__file__).with_name("serve").resolve())

# Serve directory for JS/CSS files
serve = {"__trame_simput": serve_path}

# List of JS files to load (usually from the serve path above)
scripts = ["__trame_simput/vue-trame_simput.umd.min.js"]

# List of Vue plugins to install/load
vue_use = ["trame_simput"]


def setup(server, **kwargs):
    server.add_protocol_to_configure(
        lambda root_protocol: root_protocol.registerLinkProtocol(SimputProtocol())
    )



def extract_ui(yaml_content):
    ui_map = {}
    prop_indent = 10
    current_type = None
    current_list = []
    last_property = None
    for line in yaml_content.splitlines():
        stline = line.strip()
        sline = stline.split(":")[0]
        # Skip empty lines or comments
        if len(sline) == 0 or sline[0] == "#":
            # print(f"skip a: {sline}")
            continue

        indent = line.index(sline)
        if indent == 0:
            # Detect new object definition
            if current_type:
                # print(f"Add ui for {current_type}")
                current_list.insert(0, f'<ui id="{current_type}">')
                current_list.append("</ui>")
                ui_map[current_type] = "\n".join(current_list)
                current_list = []
            current_type = line.strip().split(":")[0]
        else:
            # Detect property
            if prop_indent > indent:
                prop_indent = indent

            # skip hidden prop
            if stline == "_ui: skip":
                last_property = None
                current_list.pop()
                continue

            # skip object prop
            if stline == "_ui: proxy":
                current_list.pop()
                continue

            if stline == "type: proxy" and last_property:
                current_list.append(f'  <proxy name="{last_property}" />')

            if indent > prop_indent or sline[0] == "_":
                continue

            current_list.append(f'  <input name="{sline}" />')
            last_property = sline

    # Always have a ui container
    # print(f"Add ui for {current_type}")
    current_list.insert(0, f'<ui id="{current_type}">')
    current_list.append("</ui>")
    ui_map[current_type] = "\n".join(current_list)

    return ui_map



from pathlib import Path
from .. import utils
from .utils import extract_ui
import yaml
import xml.etree.ElementTree as ET
import logging

logger = logging.getLogger("simput.core.ui")
logger.setLevel(logging.ERROR)


class UIManager:
    """
    UI Manager provide UI information to edit and input object properties

    A UIManager is responsible to map a UI to proxy properties with the help
    of a resolver which is specialized to the target environment (Qt, Web)
    """

    id_generator = utils.create_id_generator()

    def __init__(self, proxymanager, ui_resolver):
        self._id = next(UIManager.id_generator)
        self._pxm = proxymanager
        self._ui_resolver = ui_resolver
        self._ui_xml = {}
        self._ui_lang = {}
        self._ui_resolved = {}
        # event handling
        self._listeners = set()

    @property
    def id(self):
        """Return Manager id"""
        return f"{self._pxm.id}:{self._id}"

    @property
    def proxymanager(self):
        """Return linked proxy manager"""
        return self._pxm

    def clear_ui(self):
        """Clear any loaded UI definition"""
        self._ui_xml = {}
        self._ui_resolved = {}

    # -------------------------------------------------------------------------
    # Definition handling
    # -------------------------------------------------------------------------

    def load_model(self, yaml_file=None, yaml_content=None):
        self.load_language(yaml_file=yaml_file, yaml_content=yaml_content)
        return self.proxymanager.load_model(yaml_file, yaml_content)

    def load_language(self, yaml_file=None, yaml_content=None, clear_ui=False):
        """Load language for the objects form"""
        if clear_ui:
            self.clear_ui()

        if yaml_file:
            path = Path(yaml_file)
            if path.exists():
                yaml_content = path.read_text(encoding="UTF-8")

        if yaml_content:
            self._ui_lang.update(yaml.safe_load(yaml_content))
            auto_ui = extract_ui(yaml_content)
            self._ui_resolved = {}
            ui_change_count = 0
            for ui_type in auto_ui:
                if ui_type not in self._ui_xml:
                    self._ui_xml[ui_type] = auto_ui[ui_type]
                    ui_change_count += 1

            if ui_change_count:
                self._emit("lang+ui")
            else:
                self._emit("lang")

            return True

        return False

    def load_ui(self, xml_file=None, xml_content=None, clear_ui=False):
        """Load layout for the objects form"""
        if clear_ui:
            self.clear_ui()

        if xml_file:
            path = Path(xml_file)
            if path.exists():
                xml_content = path.read_text(encoding="UTF-8")

        if xml_content:
            root = ET.fromstring(xml_content)
            for child in root:
                obj_type = child.attrib["id"]
                self._ui_xml[obj_type] = ET.tostring(child).decode("UTF-8").strip()

            self._ui_resolved = {}
            self._emit("ui")
            return True

        return False

    # -------------------------------------------------------------------------
    # Event handling
    # -------------------------------------------------------------------------

    def _emit(self, topic, **kwargs):
        for listener in self._listeners:
            listener(topic, **kwargs)

    def on(self, fn_callback):
        """
        Register callback when something is changing in ObjectManager.

        fn(topic, **kwars)
        => topic='ui'
        => topic='lang'
        => topic='lang+ui'
        """
        self._listeners.add(fn_callback)

    def off(self, fn_callback):
        """
        Unregister callback
        """
        self._listeners.discard(fn_callback)

    # -------------------------------------------------------------------------
    # UI handling
    # -------------------------------------------------------------------------

    def data(self, proxy_id):
        """Return proxy state to fill UI with"""
        _proxy = self._pxm.get(proxy_id)
        if _proxy:
            return _proxy.state

        logger.info("UIManager::data(%s) => No proxy", proxy_id)
        return None

    def ui(self, _type):
        """Return resolved layout"""
        if _type in self._ui_resolved:
            return self._ui_resolved[_type]

        model_def = self._pxm.get_definition(_type)
        lang_def = self._ui_lang[_type]
        ui_def = self._ui_xml[_type]
        resolved = self._ui_resolver.resolve(model_def, lang_def, ui_def)
        self._ui_resolved[_type] = resolved

        return resolved



from trame_client.widgets.core import AbstractElement
from .. import module
from ..module.core import SimputController


class HtmlElement(AbstractElement):
    def __init__(self, _elem_name, children=None, **kwargs):
        super().__init__(_elem_name, children, **kwargs)
        if self.server:
            self.server.enable_module(module)


class Simput(HtmlElement):
    """
    Simput data management component. This must be set as the root of a layout to provide children with Simput data.

    :param ui_manager: See simput docs |simput_link| for more info
    :param domains_manager: See simput docs |simput_link| for more info
    :param prefix: Constructing a Simput component will set several variables, optionally prefixed by a namespace
    :type prefix:  str | None
    :param query: String filtering
    :type query: str
    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None

    >>> layout.root = simput.Simput(ui_manager, prefix="myForm")
    """

    def __init__(self, ui_manager, prefix=None, children=None, **kwargs):
        super().__init__("Simput", children, **kwargs)
        ns = f"simput_{self._id}"
        if prefix:
            ns = prefix
        self._ref = kwargs.get("ref", ns)
        self._helper = SimputController(self._server, ui_manager, namespace=ns)
        self._attributes["namespace"] = f'namespace="{ns}"'
        self._attributes["ref"] = f'ref="{self._ref}"'
        self._attr_names += ["query"]

    @property
    def helper(self):
        """
        Simput helper object
        """
        return self._helper

    def apply(self, **kwargs):
        """
        Flush modified properties so they can be pushed to their concrete objects
        """
        self._helper.apply()

    def reset(self, **kwargs):
        """
        Unapply properties
        """
        self._helper.reset()

    def push(self, id=None, type=None, domains=None, proxy=None, **kwargs):
        """
        Ask server to push data, ui, or constraints
        """
        if proxy is not None:
            id = proxy
            domains = proxy
        self._helper.push(id=id, type=type, domains=domains)

    def update(self, change_set, **kwargs):
        """
        List of properties and value to update

        >>> change_set = [
        ... {"id":"12", "name":"Radius", "value": 0.75},
        ... {"id": "12", "name":"Resolution", "value": 24}
        ... ]

        """
        self._helper.update(change_set)

    def refresh(self, id=0, property="", **kwargs):
        self._helper.refresh(id, property)

    def reload(self, name):
        self._server.js_call(self._ref, "reload", name)

    def reload_domain(self):
        self.reload("domain")

    def reload_data(self):
        self.reload("data")

    def reload_ui(self):
        self.reload("ui")

    @property
    def changeset(self):
        """
        All unapplied changesets
        """
        return self._helper.changeset

    @property
    def has_changes(self):
        """
        Does the changeset have content?
        """
        return self._helper.has_changes

    @property
    def auto_update(self):
        """
        Whether to automatically apply changes
        """
        return self._helper.auto_update

    @auto_update.setter
    def auto_update(self, value):
        self._helper.auto_update = value


class SimputItem(HtmlElement):
    """
    Simput data display component. This must be child of a Simput component to have access to Simput data. See simput docs |simput_link| for more info.

    :param item_id: The simput id of the data to display
    :type item_id: str
    :param extract: Columns to make available from this component to its children
    :type extract: list[str]
    :param no_ui: Whether to show simput template UI
    :type no_ui: bool
    :param v_slot: Fields to be pass to slot template (i.e.: data, ui, domain, properties, all...)
    :type v_slot: str
    :param children: The children nested within this element
    :type children:  str | list[trame.html.*] | trame.html.* | None

    Events

    :param dirty: Function to call when itemId is changed
    :type dirty: function
    """

    def __init__(self, children=None, extract=[], **kwargs):
        super().__init__("SimputItem", children, **kwargs)
        self._attr_names += [
            ("item_id", ":itemId"),
            "no_ui",
            ("v_slot", "v-slot"),
        ]
        self._event_names += [
            "dirty",
        ]

        if extract:
            self._attributes["prop_extract"] = f'#properties="{{{", ".join(extract)}}}"'


__all__ = [
    "Simput",
    "SimputItem",
]



import logging
import yaml
import json
from pathlib import Path
from . import mapping, domains, utils


logger = logging.getLogger("simput.core.proxy")
logger.setLevel(logging.WARN)

# -----------------------------------------------------------------------------
# Proxy
# -----------------------------------------------------------------------------


class Proxy:
    """
    A Proxy keep track of a set of a properties for an other object.
    Proxy can flush its local state to the object it controls by calling commit().
    To reset uncommitted changes, just call reset().
    Proxy properties can be access with the . and [] notation.
    Proxy have states that are easily serializable.
    """

    __id_generator = utils.create_id_generator()
    __api = set(
        [
            "definition",
            "id",
            "type",
            "object",
            "manager",
            "modified",
            "mtime",
            "edited_property_names",
            "tags",
            "own",
            "set_property",
            "get_property",
            "list_property_names",
            "commit",
            "reset",
            "on",
            "off",
            "state",
            "remap_ids",
        ]
    )

    def __init__(
        self,
        __proxy_manager,
        __type,
        __object=None,
        _proxy_id=None,
        _name=None,
        _tags=[],
        _object_adapter=None,
        skip_object_init=False,
        **kwargs,
    ):
        self._id = _proxy_id or next(Proxy.__id_generator)
        self._name = _name or __type
        self._mtime = __proxy_manager.mtime
        self._proxy_manager = __proxy_manager
        self._type = __type
        self._pushed_properties = {}
        self._properties = {}
        self._dirty_properties = set()
        self._listeners = set()
        self._tags = set(_tags)
        self._tags.update(self.definition.get("_tags", []))
        self._object_adapter = _object_adapter
        self._domains = {}

        if self._object_adapter is None:
            self._object_adapter = mapping.get_default_object_adapter()

        # Proxy can be fully virtual (:None)
        if skip_object_init:
            self._object = None
        else:
            self._object = __object

        # proxy id that we created and therefore that we should manage
        self._own = set()

        # Handle registration
        self._proxy_manager._id_map[self._id] = self
        for tag in self._tags:
            self._proxy_manager._tag_map.setdefault(tag, set()).add(self._id)

        # handle initial
        for _prop_name, _prop_def in self.definition.items():
            if _prop_name.startswith("_"):
                continue

            _init_def = _prop_def.get("initial", None)
            if _prop_name in kwargs:
                self.set_property(_prop_name, kwargs[_prop_name])
            elif isinstance(_init_def, dict):
                logger.error("Don't know how to deal with domain yet: %s", _init_def)
            else:
                self.set_property(_prop_name, _init_def)

        # handle domains
        for _prop_name, _prop_def in self.definition.items():
            if _prop_name.startswith("_"):
                continue
            _prop_domains = self._domains.setdefault(_prop_name, {})
            for domain_def in _prop_def.get("domains", []):
                _type = domain_def.get("type")
                _name = domain_def.get("name", _type)
                _prop_domain = domains.create_property_domain(
                    self, _prop_name, **domain_def
                )

                if _prop_domain is None:
                    logger.warn("No matching domain for type(%s)", _type)
                    continue

                if not _prop_domain:
                    continue

                if _name not in _prop_domains:
                    _prop_domains[_name] = _prop_domain
                else:
                    # Name conflict on domain/prop
                    count = 1
                    while f"{_name}_{count}" in _prop_domains:
                        count += 1
                    _prop_domains[f"{_name}_{count}"] = _prop_domain

                # Try default set
                _prop_domain.set_value()

        # May need several pass
        while self.domains_apply():
            pass

        self._object = __object

    def __del__(self):
        if self._object_adapter:
            self._object_adapter.before_delete(self)
        logger.info("Proxy deleted %s[%s]", self.type, self.id)

    @property
    def definition(self):
        """Return Proxy definition"""
        return self._proxy_manager.get_definition(self._type)

    @property
    def id(self):
        """Return Proxy ID"""
        return self._id

    @property
    def type(self):
        """Return Proxy Type"""
        return self._type

    @property
    def object(self):
        """Return Proxy concrete object if any"""
        return self._object

    @property
    def manager(self):
        """Return ProxyManager that owns us"""
        return self._proxy_manager

    def modified(self):
        """Mark proxy modified"""
        self._mtime = self.manager.modified()

    @property
    def mtime(self):
        """Return proxy modified time"""
        return self._mtime

    @property
    def edited_property_names(self):
        """Return the list of properties that needs to be pushed"""
        return self._dirty_properties

    @property
    def tags(self):
        """Return the list of tags of that proxy"""
        return self._tags

    @tags.setter
    def tags(self, value):
        """Update proxy tag"""
        self._tags = set(value)

    @property
    def own(self):
        """List of proxy ids we created"""
        return self._own

    @own.setter
    def own(self, ids):
        """Update list of proxy we own"""
        if isinstance(ids, str):
            # single id
            self._own.add(ids)
        elif isinstance(ids, Proxy):
            self._own.add(ids.id)
        else:
            self._own.update(ids)

    def set_property(self, name, value):
        """Update a property on that proxy"""
        # convert any invalid indirect value (proxy)
        definition = self.definition.get(name, None)
        if definition is None:
            logger.warn("No definition found for '%s'", name)
            return False
        prop_type = definition.get("type", "string")
        safe_value = value
        if value is not None:
            if prop_type == "proxy" and not isinstance(value, str):
                safe_value = value.id

        # check if change
        change_detected = False
        prev_value = self._properties.get(name, None)
        saved_value = self._pushed_properties.get(name, None)
        if utils.is_equal(safe_value, saved_value):
            self._dirty_properties.discard(name)
        elif not utils.is_equal(safe_value, prev_value):
            self._dirty_properties.add(name)
            change_detected = True
        self._properties[name] = safe_value

        if change_detected:
            self._proxy_manager.dirty_proxy(self._id)

        if self._object:
            self._object_adapter.update(self, name)

        self._emit(
            "update",
            modified=change_detected,
            property_name=name,
            properties_dirty=list(self._dirty_properties),
        )

        return change_detected

    def get_property(self, name, default=None):
        """Return a property value"""
        value = self._properties.get(name, default)
        if "proxy" == self.definition.get(name).get("type"):
            return self._proxy_manager.get(self._properties.get(name))

        return value

    def list_property_names(self):
        """Return the list of property names"""
        return [name for name in self.definition if not name.startswith("_")]

    def commit(self):
        """Flush modified properties"""
        self._proxy_manager.clean_proxy_data(self._id)
        if self._dirty_properties:
            properties_dirty = list(self._dirty_properties)
            if self._object:
                self._object_adapter.commit(self)

            self._pushed_properties.update(self._properties)
            self._dirty_properties.clear()

            for _sub_id in self.own:
                self._proxy_manager.get(_sub_id).commit()

            self._emit("commit", properties_dirty=properties_dirty)
            return True
        return False

    def reset(self):
        """Undo any uncommitted properties"""
        self._proxy_manager.clean_proxy_data(self._id)
        if self._dirty_properties:
            properties_dirty = list(self._dirty_properties)
            self._dirty_properties.clear()
            self._properties.update(self._pushed_properties)

            if self._object:
                self._object_adapter.reset(self, properties_dirty)

            self._emit("reset", properties_dirty=properties_dirty)
            return True

    def fetch(self):
        self._object_adapter.fetch(self)

    def on(self, fn):
        """
        Register listener:
        fn(topic, **kwars)
        => topic='reset' | properties_dirty=[]
        => topic='commit' | properties_dirty=[]
        => topic='update' | modified=bool, properties_dirty=[], properties_change=[]
        """
        self._listeners.add(fn)

    def off(self, fn):
        """Unegister listener"""
        self._listeners.discard(fn)

    def _emit(self, topic, *args, **kwargs):
        for fn in self._listeners:
            try:
                fn(topic, *args, **kwargs)
            except Exception:
                print(f"Error calling {fn} for {topic}:{args}, {kwargs}")

    def __getitem__(self, name):
        """value = proxy[prop_name]"""

        if self._properties and name in self._properties:
            if "proxy" == self.definition.get(name).get("type"):
                return self._proxy_manager.get(self._properties.get(name))

            return self._properties[name]

        logger.error("Proxy[%s] not found", name)

        raise AttributeError()

    def __setitem__(self, name, value):
        """proxy[prop_name] = value"""
        if name in self._properties and self.set_property(name, value):
            self._emit(
                "update",
                modified=True,
                properties_dirty=list(self._dirty_properties),
                properties_change=[name],
            )
        else:
            logger.error("%s.%s is not defined", name, self._elem_name)

    def __getattr__(self, name: str):
        """value = proxy.prop_name"""
        if name.startswith("_"):
            return self.__dict__.get(name, None)

        if name in Proxy.__api:
            return self.__dict__.get(name, None)

        # Fallback to properties
        return self.__getitem__(name)

    def __setattr__(self, name: str, value):
        """proxy.prop_name = value"""
        if name.startswith("_"):
            super().__setattr__(name, value)

        if self._properties and name in self._properties:
            self.__setitem__(name, value)
        else:
            super().__setattr__(name, value)

    @property
    def state(self):
        """Return proxy state that is easily serializable"""
        _properties = {}
        _obj_def = self.definition

        for prop_name, prop_def in _obj_def.items():
            if prop_name.startswith("_"):
                continue

            _properties[prop_name] = self._properties.get(prop_name, None)

        return {
            "id": self._id,
            "type": self._type,
            "name": self._name,
            "tags": list(self._tags),
            "mtime": self._mtime,
            "own": list(self._own),
            "properties": _properties,
        }

    @state.setter
    def state(self, value):
        """Use to rebuild a proxy state from an exported state"""
        self._own = set(value.get("own", []))
        self._tags.update(value.get("tags", []))
        for prop_name, prop_value in value.get("properties", {}).items():
            self.set_property(prop_name, prop_value)

        for tag in self._tags:
            self._proxy_manager._tag_map.setdefault(tag, set()).add(self._id)

    def remap_ids(self, id_map):
        """Use to remap id when reloading an exported state"""
        # Update proxy dependency
        _new_own = set()
        for old_id in self._own:
            _new_own.add(id_map[old_id])
        self._own = _new_own

        # Update proxy props
        for prop_name, prop_def in self.definition.items():
            if prop_name.startswith("_"):
                continue
            if prop_def.get("type", "") == "proxy":
                self._properties[prop_name] = id_map[self._properties[prop_name]]

    # domain api --------------------------------------------------------------

    def get_property_domains(self, property_name):
        """Helper to get the map of domains linked to a property"""
        return self._domains.get(property_name, {})

    def domains_apply(self, *property_names):
        """
        Ask domains to set values or just for one property if property_name is provided.
        Return the number of properties that have been updated.
        """
        change_count = 0
        selection = self._domains

        # Filter execution scope
        if property_names:
            selection = {}
            for name in property_names:
                selection[name] = self._domains.get(name, {})

        # Evaluate domains on selection
        for prop_domains in selection.values():
            for domain in prop_domains.values():
                if domain.set_value():
                    change_count += 1

        return change_count

    @property
    def domains_state(self):
        """
        Return a serializable state of the domains associated to this proxy.
        This include for each property and each domain a `valid` and `available` property.
        Also at the property level a list of `hints`.

        ```
        state = {
            ContourBy: {
                FieldSelector: {
                    valid: True,
                    available: [
                        { text: "Temperature", value: "Point::Temperature", ... },
                        ...
                    ]
                },
                hints: [],
            },
            Scalar: {
                Range: {
                    valid: True,
                    available: [0.5, 123.5],
                },
                hints: [
                    { level: 0, message: "Outside of range (0.5, 123.5)" },
                ],
            },
        }
        ```
        """
        output = {}
        for prop_name, prop_domains in self._domains.items():
            prop_info = {}
            hints = []

            for domain_name, domain_inst in prop_domains.items():
                available = domain_inst.available()
                valid = domain_inst.valid()
                hints += domain_inst.hints()
                if available or not valid:
                    prop_info[domain_name] = {"available": available, "valid": valid}

            if prop_info or hints:
                prop_info["hints"] = hints
                output[prop_name] = prop_info

        return output


# -----------------------------------------------------------------------------
# ProxyManager
# -----------------------------------------------------------------------------
class ProxyManager:
    """
    A ProxyManager needs to load some definitions in order to be able to create
    proxies which will hold a set of values in their properties.
    A proxy state can then be used to control a concrete object that can be
    local or remote.
    Proxies provide a nice infrastructure to map a UI to their state with
    domains and more.
    The ProxyManager is responsible for keeping track of proxies lifecycle and
    finding them from their Id or Tags.
    """

    __id_generator = utils.create_id_generator("pxm_")

    def __init__(self, object_factory=None, object_adapter=None):
        self._id = next(ProxyManager.__id_generator)

        self._mtime = 1
        self._listeners = set()
        self._life_cycle_listeners = set()
        self._obj_factory = object_factory
        self._obj_adapter = object_adapter

        self._model_definition = {}
        self._id_map = {}
        self._tag_map = {}
        self.dirty_proxy_data = set()
        self.dirty_proxy_domains = set()

    @property
    def id(self):
        """Return manager id"""
        return self._id

    @property
    def mtime(self):
        """Return current global modified time"""
        return self._mtime

    def modified(self):
        """Create a modified event and bump global mtime"""
        self._life_cycle("before_modified", mtime=self._mtime)
        self._mtime += 1
        self._life_cycle("after_modified", mtime=self._mtime)
        return self._mtime

    def _apply_mixin(self, *names):
        """Internal helper to decorate definition using some composition logic"""
        if len(names) == 0:
            names = self._model_definition.keys()

        for name in names:
            if name.startswith("_"):
                continue
            object_definition = self._model_definition.get(name, {})
            mixins = object_definition.get("_mixins", [])
            for mixin_name in mixins:
                mixin = self._model_definition.get(mixin_name, {})
                object_definition.update(mixin)

    # -------------------------------------------------------------------------
    # Event handling
    # -------------------------------------------------------------------------

    def _life_cycle(self, cycle, **kwargs):
        """Call lyfe cycle listeners"""
        for listener in self._life_cycle_listeners:
            getattr(listener, cycle)(**kwargs)

    def add_life_cycle_listener(self, listener):
        """Register life cycle listener"""
        listener.set_proxymanager(self)
        self._life_cycle_listeners.add(listener)

    def remove_life_cycle_listener(self, listener):
        """Unregister life cycle listener"""
        listener.set_proxymanager(None)
        self._life_cycle_listeners.discard(listener)

    def _emit(self, topic, **kwargs):
        for listener in self._listeners:
            listener(topic, **kwargs)

    def on(self, fn_callback):
        """
        Register callback when something is changing in ProxyManager.

        fn(topic, **kwars)
        => topic='created' | ids=[]
        => topic='changed' | ids=[]
        => topic='deleted' | ids=[]
        => topic='commit' | ids=[]
        => topic='reset' | ids=[]
        """
        self._listeners.add(fn_callback)

    def off(self, fn_callback):
        """
        Unregister attached function/method
        """
        self._listeners.discard(fn_callback)

    # -------------------------------------------------------------------------
    # Definition handling
    # -------------------------------------------------------------------------

    def load_model(self, yaml_file=None, yaml_content=None):
        """Load Data Model from YAML definition"""
        if yaml_file:
            path = Path(yaml_file)
            if path.exists():
                yaml_content = path.read_text(encoding="UTF-8")

        if yaml_content:
            add_on_dict = yaml.safe_load(yaml_content)
            self._life_cycle("before_load_model", definition=add_on_dict)
            self._model_definition.update(add_on_dict)
            self._apply_mixin(*add_on_dict.keys())
            self._life_cycle("after_load_model", definition=yaml_content)
            return True

        return False

    def get_definition(self, obj_type):
        """Return a loaded definition for a given object_type"""
        return self._model_definition.get(obj_type)

    def types(self, *with_tags):
        """List proxy_types from definition that has the set of provided tags"""
        result = []
        tag_filter = set(with_tags)
        for type_name in self._model_definition.keys():
            has_tag = set(self._model_definition[type_name].get("_tags", []))
            if tag_filter.issubset(has_tag):
                result.append(type_name)

        return result

    # -------------------------------------------------------------------------
    # Proxy management
    # -------------------------------------------------------------------------

    def create(self, proxy_type, proxy_id=None, existing_obj=None, **initial_values):
        """
        Create a new instance of a proxy using a proxy_type.
        Optionally, you can provide a proxy_id to override the id of the created
        proxy instance.
        And you can provide a set of property values that will be pre-initialized
        using the **kwargs approach.
        """

        # Can't create object if no definition available
        if proxy_type not in self._model_definition:
            raise ValueError(
                f"Object of type: {proxy_type} was not found in our loaded model definitions"
            )

        self._life_cycle(
            "proxy_create_before", proxy_type=proxy_type, initial_values=initial_values
        )

        if existing_obj is not None:
            obj = existing_obj
        else:
            obj = self._obj_factory.create(proxy_type) if self._obj_factory else None

        proxy = Proxy(
            self,
            proxy_type,
            obj,
            skip_object_init=existing_obj is not None,
            **{
                "_proxy_id": proxy_id,
                "_object_adapter": self._obj_adapter,
                **initial_values,
            },
        )
        self._life_cycle(
            "proxy_create_before_commit",
            proxy_type=proxy_type,
            initial_values=initial_values,
            proxy=proxy,
        )

        if existing_obj is not None:
            proxy.fetch()
        else:
            proxy.commit()

        self._life_cycle(
            "proxy_create_after_commit",
            proxy_type=proxy_type,
            initial_values=initial_values,
            proxy=proxy,
        )

        self._emit("created", ids=[proxy.id])

        return proxy

    def delete(self, proxy_id, trigger_modified=True):
        """
        Delete object along with its dependency that it is owner of
        """
        self._life_cycle(
            "proxy_delete_before", proxy_id=proxy_id, trigger_modified=trigger_modified
        )
        before_delete = set(self._id_map.keys())
        # Delete ourself
        proxy_to_delete: Proxy = self._id_map[proxy_id]
        del self._id_map[proxy_id]

        for tag in proxy_to_delete.tags:
            self._tag_map.get(tag).discard(proxy_id)

        self._life_cycle(
            "proxy_delete_after_self",
            proxy_id=proxy_id,
            trigger_modified=trigger_modified,
            proxy=proxy_to_delete,
        )

        # Delete objects that we own
        for _id in proxy_to_delete.own:
            self.delete(_id, False)

        self._life_cycle(
            "proxy_delete_after_own",
            proxy_id=proxy_id,
            trigger_modified=trigger_modified,
            proxy=proxy_to_delete,
        )

        if trigger_modified:
            after_delete = set(self._id_map.keys())
            self.modified()
            self._emit("deleted", ids=list(before_delete.difference(after_delete)))

    def get(self, proxy_id: str) -> Proxy:
        """
        return proxy instance
        """
        return self._id_map.get(proxy_id, None)

    def update(self, change_set):
        """
        changeSet = [
            { id: 2143, name: 'Origin', value: [0, 1, 3] },
            ...
        ]
        """
        self._life_cycle("proxy_update_before", change_set=change_set)
        dirty_ids = set()
        dirty_proxies_to_commit = set()
        for change in change_set:
            _id = change["id"]
            _name = change["name"]
            _value = change["value"]
            dirty_ids.add(_id)
            proxy: Proxy = self.get(_id)
            if "auto_commit" in proxy.tags:
                dirty_proxies_to_commit.add(proxy)
            proxy.set_property(_name, _value)

        # commit changes for proxy tagged as auto_commit
        for proxy in dirty_proxies_to_commit:
            proxy.commit()
            self._emit("commit", ids=[proxy.id])

        self._life_cycle(
            "proxy_update_after", change_set=change_set, dirty_ids=dirty_ids
        )
        self._emit("changed", ids=list(dirty_ids))

    def get_instances_of_type(self, proxy_type):
        """
        Return all the instances of the given type
        """
        result = []
        for proxy in self._id_map.values():
            if proxy.type == proxy_type:
                result.append(proxy)

        return result

    def tags(self, *args):
        """List all instances containing all the listed tags"""
        selected_ids = set(self._id_map.keys())
        for tag in args:
            if tag in self._tag_map:
                selected_ids &= self._tag_map[tag]
            else:
                return []

        result = []
        for obj_id in selected_ids:
            result.append(self._id_map[obj_id])

        return result

    # -------------------------------------------------------------------------
    # Import / Export
    # -------------------------------------------------------------------------

    def save(self, file_output=None):
        """Export state (definition+data) into a file"""
        self._life_cycle("export_before", file_output=file_output)
        data = {
            "model": self._model_definition,
            "proxies": [proxy.state for proxy in self._id_map.values()],
        }
        self._life_cycle("export_after", file_output=file_output, data=data)
        if file_output:
            with open(file_output, "w") as outfile:
                json.dump(data, outfile)
        else:
            return json.dumps(data)

    def load(self, file_input=None, file_content=None):
        """Load previously exported state from a file"""
        self._life_cycle(
            "import_before", file_input=file_input, file_content=file_content
        )
        if file_input:
            with open(file_input) as json_file:
                data = json.load(json_file)
        else:
            data = json.loads(file_content)

        self._life_cycle(
            "import_before_processing",
            file_input=file_input,
            file_content=file_content,
            data=data,
        )

        self._model_definition.update(data["model"])

        # Create proxies
        _id_remap = {}
        _new_ids = []
        for proxy_state in data["proxies"]:
            _id = proxy_state["id"]
            _type = proxy_state["type"]
            _proxy = self.create(_type)
            _id_remap[_id] = _proxy.id
            _proxy.state = proxy_state
            _new_ids.append(_proxy.id)

        # Remap ids
        for new_id in _new_ids:
            _proxy = self.get(new_id)
            _proxy.remap_ids(_id_remap)
            _proxy.commit()

        self._life_cycle(
            "import_after",
            file_input=file_input,
            file_content=file_content,
            data=data,
            new_ids=_new_ids,
            id_remap=_id_remap,
        )
        self._emit("created", ids=_new_ids)

    # -------------------------------------------------------------------------
    # Commit / Reset
    # -------------------------------------------------------------------------

    def commit_all(self):
        """Commit all dirty proxies"""
        proxy_ids = list(self.dirty_proxy_data)
        for _id in proxy_ids:
            proxy = self.get(_id)
            if proxy:
                proxy.commit()
        self._emit("commit", ids=proxy_ids)

    def reset_all(self):
        """Reset all dirty proxies"""
        proxy_ids = list(self.dirty_proxy_data)
        for _id in proxy_ids:
            proxy = self.get(_id)
            if proxy:
                proxy.reset()
        self._emit("reset", ids=proxy_ids)

    # -------------------------------------------------------------------------
    # Dirty / Clean management
    # -------------------------------------------------------------------------

    def dirty_proxy(self, id):
        self.dirty_proxy_data.add(id)
        self.dirty_proxy_domains.add(id)

    def clean_proxy_data(self, id=None):
        if id is None:
            self.dirty_proxy_data.clear()
        else:
            self.dirty_proxy_data.discard(id)

    def clean_proxy_domains(self, id):
        if id is None:
            self.dirty_proxy_domains.clear()
        else:
            self.dirty_proxy_domains.discard(id)

    def list_and_clean_proxy_data(self):
        ids = list(self.dirty_proxy_data)
        self.dirty_proxy_data.clear()
        return ids

    def list_and_clean_proxy_domains(self):
        ids = list(self.dirty_proxy_domains)
        self.dirty_proxy_domains.clear()
        return ids



import os
import json
import logging
from pathlib import Path

logger = logging.getLogger("simput.core.controller")
logger.setLevel(logging.ERROR)


class SimputController:
    def __init__(
        self,
        server,
        ui_manager,
        namespace="simput",
        log_dir=None,
    ):
        logger.info("created")
        self._server = server
        self._ui_manager = ui_manager
        self._namespace = namespace
        self._pending_changeset = {}
        self._auto_update = False
        self._log_directory = log_dir if log_dir else os.environ.get("SIMPUT_LOG_DIR")

        # init keys
        self.id_key = f"{namespace}Id"
        self.changecount_key = f"{namespace}ChangeSet"
        self.changeset_key = f"{namespace}ChangeSetContent"
        self.auto_key = f"{namespace}AutoApply"
        self.apply_key = f"{namespace}Apply"
        self.reset_key = f"{namespace}Reset"
        self.fetch_key = f"{namespace}Fetch"
        self.update_key = f"{namespace}Update"
        self.refresh_key = f"{namespace}Refresh"
        self.reset_cache_key = f"{namespace}ResetCache"

        # Attach annotations
        self._server.state[self.id_key] = self._ui_manager.id
        self._server.state[self.changecount_key] = len(self.changeset)
        self._server.state[self.changeset_key] = self.changeset
        self._server.state[self.auto_key] = self._auto_update
        self._server.change(self.auto_key)(self._update_auto)
        self._server.trigger(self.apply_key)(self.apply)
        self._server.trigger(self.reset_key)(self.reset)
        self._server.trigger(self.fetch_key)(self.push)
        self._server.trigger(self.update_key)(self.update)
        self._server.trigger(self.refresh_key)(self.refresh)
        self._server.trigger(self.reset_cache_key)(self.reset_cache)

        # Monitor ui change
        self._ui_manager.on(self._ui_change)
        self._ui_manager.proxymanager.on(self._data_change)

        # Debug
        if self._log_directory:
            os.makedirs(self._log_directory, exist_ok=True)

    def __del__(self):
        logger.info("deleted")
        self._ui_manager.off(self._ui_change)
        self._ui_manager.proxymanager.off(self._data_change)

    def _log(self, id, name, content):
        if self._log_directory:
            full_path = Path(self._log_directory) / f"{id}_{name}.json"
            with open(full_path, "w") as file:
                file.write(json.dumps(content, indent=2))

    @property
    def changeset(self):
        logger.info("changeset")
        change_set = []
        for obj_id in self._pending_changeset:
            for prop_name in self._pending_changeset[obj_id]:
                value = self._pending_changeset[obj_id][prop_name]
                change_set.append(
                    {
                        "id": obj_id,
                        "name": prop_name,
                        "value": value,
                    }
                )
        return change_set

    def apply(self):
        logger.info("apply")
        self._ui_manager.proxymanager.commit_all()

        # Make sure reset don't send things twice
        self._pending_changeset = {}
        self.reset()

    def reset(self):
        logger.info("reset")
        ids_to_update = list(self._pending_changeset.keys())
        self._pending_changeset = {}
        self._server.state[self.changecount_key] = 0
        self._server.state[self.changeset_key] = []

        self._ui_manager.proxymanager.reset_all()

        # Go ahead and push changes
        for _id in ids_to_update:
            self.push(id=_id, domains=_id)

    def refresh(self, id=0, property=""):
        logger.info("refresh")
        proxy = self._ui_manager.proxymanager.get(id)
        if not proxy:
            return

        prop_domains = proxy.get_property_domains(property).values()

        # Reset
        change_count = 0
        for domain in prop_domains:
            domain.enable_set_value()
            change_count += domain.set_value()

        if change_count:
            for _id in proxy.proxymanager.list_and_clean_proxy_domains():
                self.push(id=_id)

            if self._auto_update:
                self.apply()

    def reset_cache(self):
        logger.info("reset_cache")
        self._server.protocol_call("simput.reset.cache")

    def push(self, id=None, type=None, domains=None):
        logger.info("push id(%s) - type(%s) - domains(%s)", id, type, domains)
        if id:
            self._server.protocol_call("simput.data.get", self._ui_manager.id, id)
        if type:
            self._server.protocol_call("simput.ui.get", self._ui_manager.id, type)

        try:
            if domains and int(domains):
                self._server.protocol_call(
                    "simput.domains.get", self._ui_manager.id, domains
                )
        except ValueError:
            # Invalid domain so just skip it...
            pass

    def emit(self, topic, **kwargs):
        logger.info("emit: %s", topic)
        self._server.protocol_call("simput.push.event", topic, **kwargs)

    def update(self, change_set):
        logger.info("update")
        pxm = self._ui_manager.proxymanager
        id_map = {}
        for change in change_set:
            _id = change.get("id")
            _name = change.get("name")
            _value = change.get("value")

            if _id not in self._pending_changeset:
                self._pending_changeset[_id] = {}
            _obj_change = self._pending_changeset[_id]

            _obj_change[_name] = _value

            # debug
            if self._log_directory:
                id_map.setdefault(_id, {})
                id_map[_id][_name] = _value

        if self._log_directory:
            for _id in id_map:
                self._log(_id, "change", id_map[_id])

        self._server.state[self.changecount_key] = len(self.changeset)
        self._server.state[self.changeset_key] = self.changeset

        # Update data
        pxm.update(change_set)

        # Run domains
        change_detected = 1
        all_ids = set()
        data_ids = set()

        # Execute domains
        while change_detected:
            change_detected = 0
            m_ids = pxm.list_and_clean_proxy_domains()
            all_ids.update(m_ids)
            for _id in m_ids:
                if pxm.get(_id).domains_apply():
                    data_ids.add(_id)
                    change_detected += 1

        # Push any changed state in domains
        for _id in all_ids:
            pxm.clean_proxy_domains(_id)
            proxy = pxm.get(_id)
            _domain = proxy.domains_state
            self._log(_id, "domain", _domain)
            self._server.protocol_call(
                "simput.message.push",
                {
                    "id": _id,
                    "domains": _domain,
                },
            )

        for _id in data_ids:
            _data = self._ui_manager.data(_id)
            self._log(_id, "data", _data)
            self._server.protocol_call(
                "simput.message.push",
                {
                    "id": _id,
                    "data": _data,
                },
            )

        if self._auto_update:
            self.apply()

    @property
    def has_changes(self):
        return len(self._pending_changeset) > 0

    @property
    def auto_update(self):
        return self._auto_update

    @auto_update.setter
    def auto_update(self, value):
        logger.info("auto_update")
        self._auto_update = value
        self._server.state[self.auto_key] = value

    def _update_auto(self, **kwargs):
        if self.auto_update == self._server.state[self.auto_key]:
            return
        logger.info("_update_auto")
        value = self._server.state[self.auto_key]
        self.auto_update = value
        if value:
            self.apply()

    def _ui_change(self, *args, **kwargs):
        logger.info("_ui_change")
        self.emit("ui-change")

    def _data_change(self, action, **kwargs):
        logger.info("_data_change")
        self.emit("data-change", action=action, **kwargs)

        if action == "commit":
            _ids = kwargs.get("ids", [])
            for _id in _ids:
                self._pending_changeset.pop(_id)
            self._server.state[self.changecount_key] = len(self.changeset)
            self._server.state[self.changeset_key] = self.changeset



import os
import json
from pathlib import Path

from wslink import register as exportRpc
from wslink.websocket import LinkProtocol

from trame_simput.core.factory import get_simput_manager
import logging

logger = logging.getLogger("simput.core.protocol")
logger.setLevel(logging.ERROR)


class SimputProtocol(LinkProtocol):
    def __init__(self, log_dir=None):
        logger.info("Created")
        super().__init__()
        self._log_directory = log_dir if log_dir else os.environ.get("SIMPUT_LOG_DIR")
        self.reset_cache()

        if self._log_directory:
            os.makedirs(self._log_directory, exist_ok=True)

    def _log(self, id, name, content):
        if self._log_directory:
            full_path = Path(self._log_directory) / f"{id}_{name}.json"
            with open(full_path, "w") as file:
                file.write(json.dumps(content, indent=2))

    @exportRpc("simput.reset.cache")
    def reset_cache(self):
        logger.info("reset_cache")
        self.net_cache_domains = {}

    @exportRpc("simput.push")
    def push(self, manager_id, id=None, type=None):
        logger.info("push")
        uim = get_simput_manager(manager_id)
        message = {"id": id, "type": type}
        if id is not None:
            _data = uim.data(id)
            self._log(id, "data", _data)
            message.update({"data": _data})

        if type is not None:
            message.update({"ui": uim.ui(type)})

        self.publish("simput.push", message)

    @exportRpc("simput.data.get")
    def get_data(self, manager_id, id):
        logger.info("get_data")
        uim = get_simput_manager(manager_id)
        _data = uim.data(id)
        self._log(id, "data", _data)
        msg = {"id": id, "data": _data}
        self.send_message(msg)
        return msg

    @exportRpc("simput.ui.get")
    def get_ui(self, manager_id, type):
        logger.info("get_ui(%s)", type)
        uim = get_simput_manager(manager_id)
        msg = {"type": type, "ui": uim.ui(type)}
        self.send_message(msg)
        return msg

    @exportRpc("simput.domains.get")
    def get_domains(self, manager_id, id):
        logger.info("get_domains(%s)", id)
        msg = {"id": id, "domains": {}}

        pxm = get_simput_manager(manager_id).proxymanager
        pxm.clean_proxy_domains(id)
        _domain = pxm.get(id).domains_state
        self._log(id, "domain", _domain)
        msg["domains"] = _domain

        self.send_message(msg)
        return msg

    @exportRpc("simput.message.push")
    def send_message(self, message):
        logger.info("send_message")
        # Cache domain to prevent network call
        # when not needed. (optional)
        if "domains" in message:
            _id = message.get("id")
            content = self.net_cache_domains.get(_id)
            to_send = json.dumps(message)
            if content == to_send:
                return
            self.net_cache_domains[_id] = to_send
        # - end
        self.publish("simput.push", message)

    @exportRpc("simput.push.event")
    def emit(self, topic, **kwargs):
        logger.info("emit %s", topic)
        event = {"topic": topic, **kwargs}
        self.publish("simput.event", event)



import logging

logger = logging.getLogger("simput.core.mapping")
logger.setLevel(logging.INFO)

# ----------------------------------------------------------------------------
# Default singleton
# ----------------------------------------------------------------------------
DEFAULT_PROXY_OBJECT_ADAPTER = None


def get_default_object_adapter():
    global DEFAULT_PROXY_OBJECT_ADAPTER
    if DEFAULT_PROXY_OBJECT_ADAPTER is None:
        DEFAULT_PROXY_OBJECT_ADAPTER = ProxyObjectAdapter()

    return DEFAULT_PROXY_OBJECT_ADAPTER


# ----------------------------------------------------------------------------
# ProxyObjectAdapter
# ----------------------------------------------------------------------------
class ProxyObjectAdapter:
    """API for proxy/object synchronization"""

    @staticmethod
    def commit(proxy):
        pass

    @staticmethod
    def reset(proxy, props_to_reset=[]):
        pass

    @staticmethod
    def fetch(proxy):
        pass

    @staticmethod
    def update(proxy, *property_names):
        pass

    @staticmethod
    def before_delete(proxy):
        pass


# ----------------------------------------------------------------------------
# ObjectFactory
# ----------------------------------------------------------------------------
class ObjectFactory:
    """
    Concrete object factory

    The ObjectFactory is responsible for the creation of concrete object that
    a given proxy can control.
    """

    def __init__(self):
        self._map = {}

    def register(self, name, klass):
        """Register constructor to match definition type"""
        self._map[name] = klass

    def create(self, name, **kwargs):
        """Try to create concreate object"""
        obj = None
        if name in self._map:
            obj = self._map[name](**kwargs)

        if name in globals():
            obj = globals()[name](**kwargs)

        if obj is None:
            logger.error("Could not instantiate", name)

        return obj



import xml.etree.ElementTree as ET

VUETIFY_MAP = {
    "ui": "div",
    "row": "v-row",
    "col": "v-col",
    "spacer": "v-spacer",
    "divider": "v-divider",
    "show": "sw-show",
    "hide": "sw-hide",
    "text": "sw-text",
}

WIDGET_MAP = {
    "input": "sw-text-field",
    "textarea": "sw-text-area",
}

WIDGET_KNOWN = {
    "div",
    "v-row",
    "v-col",
    "v-spacer",
    "v-divider",
    "sw-show",
    "sw-hide",
    "sw-text",
    "sw-text-field",
    "sw-text-area",
    "sw-slider",
    "sw-group",
}


class VuetifyResolver:
    def __init__(self):
        self._model = None
        self._labels = None

    def get_widget(self, elem):
        attributes = {}
        if elem.tag in VUETIFY_MAP:
            return VUETIFY_MAP[elem.tag], attributes
        elif elem.tag == "input":
            domains = self._model.get(elem.get("name"), {}).get("domains", [])
            widget = "sw-text-field"
            for domain in domains:
                ctype = domain.get("type")
                level = domain.get("level", 0)
                widget = domain.get("widget", widget)
                attributes.update(domain.get("ui_attributes", {}))
                if ctype in ["LabelList", "HasTags"]:
                    values = domain.get("values", [])
                    prop_name = domain.get("property", None)
                    if len(values) and ctype not in ["HasTags", "ProxyBuilder"]:
                        attributes[":items"] = values

                    if prop_name and ctype != "FieldSelector":
                        attributes["itemsProperty"] = prop_name
                    widget = "sw-select"
                if ctype == "Boolean":
                    widget = "sw-switch"
                if ctype == "Range" and level == 2:
                    value_range = domain.get("value_range", None)
                    if value_range:
                        attributes[":min"] = str(value_range[0])
                        attributes[":max"] = str(value_range[1])
                    widget = "sw-slider"

                if ctype == "UI":
                    attributes.update(domain.get("properties", {}))
                    custom_widget = domain.get("widget", None)
                    if custom_widget and custom_widget in WIDGET_MAP:
                        widget = WIDGET_MAP[custom_widget]

            if widget is None:
                widget = "sw-text-field"

            if self._model.get(elem.get("name"), {}).get("type", "string") == "bool":
                widget = "sw-switch"

            return widget, attributes
        elif elem.tag == "proxy":
            return "sw-proxy", attributes

        if not (elem.tag in WIDGET_KNOWN or elem.tag.startswith("sw-")):
            print(f"Unknown widget element {elem.tag}")
        return elem.tag, attributes

    def process_node(self, in_elem):
        # xml mapping
        key = in_elem.get("name")
        elem_lan = None
        elem_label = key
        elem_help = None
        if key:
            elem_lan = self._labels.get(key)

        if elem_lan:
            elem_label = elem_lan.get("_label", key)
            elem_help = elem_lan.get("_help", None)

        widget, add_on_attrs = self.get_widget(in_elem)
        if not widget:
            return None

        # Add name, label, help, size
        out_elem = ET.Element(widget)
        out_elem.set(":mtime", "data.mtime")

        if key is not None and key in self._model:
            size = self._model.get(key).get("size", 1)
            ptype = self._model.get(key).get("type", "string")
            out_elem.set("name", key)
            out_elem.set("size", f"{size}")
            out_elem.set("type", ptype)
            out_elem.set(":initial", f"data.original['{key}']")
        if elem_label:
            out_elem.set("label", elem_label)
        if elem_help:
            out_elem.set("help", elem_help)

        # Add-on attributes
        for key in add_on_attrs:
            if isinstance(add_on_attrs[key], (bool, int)):
                prefix = ""
                if key[0] != ":":
                    prefix = ":"
                out_elem.set(f"{prefix}{key}", f"{add_on_attrs[key]}".lower())
            else:
                out_elem.set(key, add_on_attrs[key])

        # ui(attr) => html(attr)
        for key in in_elem.attrib:
            if key != "name":
                out_elem.set(key, in_elem.attrib[key])

        # process hierarchy
        for child in in_elem:
            out_child = self.process_node(child)
            if out_child is not None:
                out_elem.append(out_child)

        return out_elem

    def resolve(self, model, labels, ui):
        self._model = model
        self._labels = labels
        in_root = ET.fromstringlist(ui)
        out_root = self.process_node(in_root)
        self._model = None
        self._labels = None
        # print(ET.tostring(out_root, encoding="UTF-8"))
        return ET.tostring(out_root, encoding="utf-8").decode("utf-8")



from .proxy import ProxyManager
from .ui import UIManager, resolvers

UI_MANAGERS = {}


def get_simput_manager(
    id=None, pxm=None, resolver=None, object_factory=None, object_adapter=None
):
    global UI_MANAGERS
    if id is not None and id in UI_MANAGERS:
        return UI_MANAGERS[id]

    if pxm is None:
        pxm = ProxyManager(object_factory=object_factory, object_adapter=object_adapter)

    if resolver is None:
        resolver = resolvers.VuetifyResolver()

    ui_manager = UIManager(pxm, resolver)
    UI_MANAGERS[ui_manager.id] = ui_manager

    if id is not None:
        UI_MANAGERS[id] = ui_manager

    return ui_manager
